#include <avr/io.h>
#include <util/delay.h>
#include <stdint.h>
#include <stdio.h>
#include "sha256.h"

#define BAUD 9600
#define UBRR F_CPU/16/BAUD-1

#define MULTIPLIER 1534
#define INCREMENT 234
#define MODULUS 6231

void init_uart(){
    UBRR0H = (unsigned char)(UBRR >> 8);
    UBRR0L = (unsigned char)(UBRR);

    // Enable TX and RX
    UCSR0B = (1 << RXEN0) | (1 << TXEN0);

    // 8 bits of data, 2 stop bits, no parity
    UCSR0C = (1 << USBS0) | (3 << UCSZ00);
}

void send_uart(unsigned char data){
    while(!(UCSR0A & (1 << UDRE0)));

    UDR0 = data;
}

volatile uint32_t prev = 2;
uint32_t lcg_rand(){
    uint32_t res = (MULTIPLIER * prev + INCREMENT) % MODULUS;
    prev = res;

    return res;
}

void lcg_seed(uint32_t val){
    prev = val;
}

void init_timer0(){
    // Enable Timer0 with x8 prescaler
    TCCR0B = (1 << CS01);
}

void init_adc(){
    // AVCC
    ADMUX = (1 << REFS0);

    // Enable ADC, 128x prescaler
    ADCSRA = (1 << ADEN) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);
}

uint16_t read_adc(){
    ADCSRA |= (1 << ADSC);

    while(ADCSRA & (1 << ADSC));

    return ADCL | (ADCH << 8);
}

volatile uint32_t result = 0;
uint32_t floating_entropy(){
    for(int i = 0; i < 5; i++){
        // Clear MUX3:0
        ADMUX &= ~(0b1111);

        // Select ADC pins 0 through 4
        ADMUX |= (i << MUX0);

        uint16_t val = read_adc();

        result += val;
    }

    result += TCNT0;

    return result;
}

int main(){
    init_uart();
    init_adc();
    init_timer0();

    _delay_ms(10);
    while(1){
        lcg_seed(floating_entropy());

        for(int i = 0; i < 7; i++){
            uint32_t num = lcg_rand();
            sha256_hash_t hash;

            char str[12];
            snprintf(str, 12, "%u%u", num, TCNT0);

            sha256(&hash, str, 12);

            for(int j = 0; j < SHA256_HASH_BYTES; j++){
                snprintf(str, 12, "%d: %x", j, hash[j]);

                for(int i = 0; i < 12; i++){
                    if(str[i] == '\0')
                        break;

                    send_uart(str[i]);
                }

                send_uart('\r');
                send_uart('\n');
            }
            

            _delay_ms(1000);
        }
    }
}
